;; Generated package description from bicep-ts-mode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "bicep-ts-mode" "0.1.4.0.20260808.85" "tree-sitter support for Bicep" 'nil :commit "9825f9b4eb68abbd78b3e78a9757f96b751e06c4" :authors '(("Jostein Kjønigsen" . "jostein@kjonigsen.net")) :maintainer '("Jostein Kjønigsen" . "jostein@kjonigsen.net") :keywords '("bicep" "languages" "tree-sitter") :url "https://github.com/josteink/bicep-ts-mode")
